/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware;

import java.io.BufferedReader;
import java.io.InputStreamReader;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;


import at.ac.tuwien.vitalab.middleware.discovery.ServiceDB;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceRegistry;
import at.ac.tuwien.vitalab.middleware.soap.ServiceContainer;
import at.ac.tuwien.vitalab.middleware.soap.services.CallbackService;
import at.ac.tuwien.vitalab.middleware.soap.services.DeploymentService;
import at.ac.tuwien.vitalab.middleware.soap.services.InfoService;

/**
 * Middleware main class.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class Middleware {

	private static Logger logger=LoggerFactory.getLogger(Middleware.class);
	
	private static CallbackService callbackService=null;
	private static InfoService infoService=null;
	private static DeploymentService deploymentService=null;
	
	public static final String NAMESPACE=
		"http://www.vitalab.tuwien.ac.at/vimoware/";

	private static boolean running=false;
	
	/**
	 * Sets the file name of the configuration.
	 * 
	 * @param fileName File name.
	 */
	public static void setConfigurationFileName(String fileName){
		MiddlewareConfiguration.setConfigurationFile(fileName);
	}
	
	/**
	 * Starts the middleware.
	 * 
	 * @throws Exception If errors occur.
	 */
	public synchronized static void start() throws Exception {
		
		if (running) {
			return;
		}
		
		running=true;
		
		try {
			callbackService=new CallbackService();
			infoService=new InfoService();
			deploymentService=new DeploymentService();
			
			ServiceContainer.startLocalServiceContainer();
			ServiceRegistry.startLocalServiceRegistry(infoService);

			deploymentService.redeployPersistentServices();		

			// register the info service
			deploymentService.deployTransientWS(infoService);

			// register the callback service
			deploymentService.deployTransientWS(callbackService);

			// register the deployment service
			deploymentService.deployTransientWS(deploymentService);

			logger.info("Middleware has been successfully started.");

		} catch (Exception e) {
			logger.error("Middleware could not be started. Aborting.");

			try {					
				stop();					
			} catch (Exception e2) {
				// ignore
			}
			throw e;
		}
	}
	
	/**
	 * Stops the middleware.
	 * 
	 * @throws Exception If errors occur.
	 */
	public synchronized static void stop() throws Exception {
		
		if (!running) {
			return;
		}
		
		running=false;
		
		// deregister the deployment service
		deploymentService.undeployWS(deploymentService);

		// deregister the callback service
		deploymentService.undeployWS(callbackService);

		// deregister the info service
		deploymentService.undeployWS(infoService);

		// deregister rest
		deploymentService.undeployAllServices();
		
		ServiceRegistry.closeLocalServiceRegistry();
		ServiceContainer.closeLocalServiceContainer();

		logger.info("Middleware has been successfully shut down.");
	}
	
	/**
	 * Deploys a Web Service.
	 * 
	 * @param ws Web Service.
	 */
	public synchronized static void deployService(AWebService ws) {
		
		deploymentService.deployTransientWS(ws);
	}

	/**
	 * Undeploys a Web Service.
	 * 
	 * @param ws Web Service.
	 */ 
	public synchronized static void undeployService(AWebService ws) {
		
		deploymentService.undeployWS(ws);
	}
	
	/**
	 * Returns the service database.
	 * 
	 * @return Service database.
	 */
	public synchronized static ServiceDB getServiceDatabase() {
		
		ServiceRegistry localRegistry=
			ServiceRegistry.getLocalServiceRegistry();

		if (localRegistry==null) {
			throw new IllegalStateException("Middleware is not running.");
		}

		return localRegistry.getServiceDataBase();
	}
	
	/**
	 * Returns the callback service of the middleware.
	 * 
	 * @return CallbackService.
	 */
	public static CallbackService getCallbackService() {
		return callbackService;
	}

	/**
	 * Returns the info service of the middleware.
	 * 
	 * @return InfoService.
	 */
	public static InfoService getInfoService() {
		return infoService;
	}

	/**
	 * Returns the deployment service of the middleware.
	 * 
	 * @return DeploymentService.
	 */
	public static DeploymentService getDeploymentService() {
		return deploymentService;
	}

	/**
	 * Starts the middleware (SOAP and discovery).
	 * 
	 * @param args Arguments are ignored.
	 */
	public static void main(String args[]) {
		
		try {
			
			start();
			
			// run  until 'exit' is typed into stdin
			BufferedReader reader=
				new BufferedReader(new InputStreamReader(System.in));
			
			String inputLine=null;
			while ((inputLine=reader.readLine())!=null && 
					!"exit".equals(inputLine.toLowerCase())) {
			}
			
			stop();
			
			System.exit(0);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}
	}
	
}
