/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;

import at.ac.tuwien.vitalab.middleware.util.Configuration;

import com.eaio.uuid.UUID;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;



/**
 * Utility for handling the middleware configuration file.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class MiddlewareConfiguration {

	private static String hostUUID=null;
	
	private static Logger logger=
		LoggerFactory.getLogger(MiddlewareConfiguration.class);

	private static Configuration config=
		new Configuration(System.getProperty(
				"middleware.config", "conf/middleware.conf"));

	private static final String CONFIGNAME_SOAPPORT="soap.port";
	private static final String CONFIGNAME_TRANSPORT="soap.transport";
	private static final String CONFIGNAME_DEPLOYMENT_NS=
		"soap.deployment_namespace";
	
	private static final String CONFIGNAME_SERVICEPATH="http.servicepath";
	
	private static final String CONFIGNAME_ADVPORT="adv.port";
	private static final String CONFIGNAME_ADVINTERVAL="adv.interval";
	private static final String CONFIGNAME_ADVUDPTIMEOUT="adv.udptimeout";	
	private static final String CONFIGNAME_ADVTTL="adv.ttl";

	
	private static final String CONFIGNAME_LOCALIP="net.ip";
	
	public static final String TRANSPORT_SOCKET="socket";
	public static final String TRANSPORT_JETTY="jetty";
	
	private static final Integer DEFAULT_SOAPPORT=new Integer(8080);
	private static final String DEFAULT_TRANSPORT=TRANSPORT_JETTY;
	private static final String DEFAULT_SERVICEPATH="/services";
	private static final String DEFAULT_DEPLOYMENT_NS=
		"http://www.vitalab.tuwien.ac.at/workpad/soap/deployment";
	
	private static final Integer DEFAULT_ADVPORT=new Integer(1427);
	private static final Integer DEFAULT_ADVINTERVAL=new Integer(5);
	private static final Integer DEFAULT_ADVUDPTIMEOUT=new Integer(2);	
	private static final Integer DEFAULT_ADVTTL=new Integer(4);

	/**
	 * (Re)sets all configuration to a new source.
	 * 
	 * @param confFile Configuration file name.
	 */
	public static void setConfigurationFile(String confFile) {
		config=new Configuration(confFile);
	}
	
	/**
	 * Returns the SOAP transport type.
	 * 
	 * @return Type identifier.
	 */
	public static String getSoapTransportType() {
		String result=(String)config.getValue(
				CONFIGNAME_TRANSPORT,
				String.class,
				DEFAULT_TRANSPORT);
		
		if (!result.toLowerCase().equals(
				TRANSPORT_JETTY) && 
				!result.toLowerCase().equals(
						TRANSPORT_SOCKET)
				) {
			logger.warning("Configuration entry '"+CONFIGNAME_TRANSPORT+
					"' must be either '"+TRANSPORT_JETTY+
					"' or '"+TRANSPORT_SOCKET+"'. " +
					"Assuming '"+DEFAULT_TRANSPORT+"'.");
			
			config.saveCorrectedValue(CONFIGNAME_TRANSPORT, 
					String.class, 
					DEFAULT_TRANSPORT);
			
			return DEFAULT_TRANSPORT;
		}
		
		return result;
	}
	
	/**
	 * Returns the SOAP deployment namespace.
	 * 
	 * @return Namespace.
	 */
	public static String getSoapDeploymentNS() {
		return (String)config.getValue(
				CONFIGNAME_DEPLOYMENT_NS,
				String.class,
				DEFAULT_DEPLOYMENT_NS);
	}

	/**
	 * Returns the http service path.
	 * 
	 * @return Path.
	 */
	public static String getHttpServicePath() {
		return (String)config.getValue(
				CONFIGNAME_SERVICEPATH,
				String.class,
				DEFAULT_SERVICEPATH);
	}

	/**
	 * Returns the SOAP port.
	 * 
	 * @return port.
	 */
	public static int getSoapPort() {
		int result=((Integer)config.getValue(
				CONFIGNAME_SOAPPORT,
				Integer.class,
				DEFAULT_SOAPPORT)).intValue();
		
		if (result<1 || result>65535) {
			logger.warning("Configuration entry '"+CONFIGNAME_SOAPPORT+
					"' must be between 1 and 65535. Assuming default port "+
					DEFAULT_SOAPPORT+".");
			
			config.saveCorrectedValue(CONFIGNAME_SOAPPORT, 
					Integer.class, 
					DEFAULT_SOAPPORT);
			
			return DEFAULT_SOAPPORT.intValue();
		}
		
		return result;
	}
	
	/**
	 * Returns the advertisment port.
	 * 
	 * @return port.
	 */
	public static int getAdvertisementPort() {
		int result=((Integer)config.getValue(
				CONFIGNAME_ADVPORT,
				Integer.class,
				DEFAULT_ADVPORT)).intValue();
		
		if (result<1 || result>65535) {
			logger.warning("Configuration entry '"+CONFIGNAME_ADVPORT+"' must "+
					"be between 1 and 65535. Assuming default port "+
					DEFAULT_ADVPORT+".");
			
			config.saveCorrectedValue(CONFIGNAME_ADVPORT, 
					Integer.class, 
					DEFAULT_ADVPORT);
			
			return DEFAULT_ADVPORT.intValue();
		}
		
		return result;
	}
	
	/**
	 * Returns the advertisment interval.
	 * 
	 * @return interval.
	 */
	public static int getAdvertisementInterval() {
		int result=((Integer)config.getValue(
				CONFIGNAME_ADVINTERVAL,
				Integer.class,
				DEFAULT_ADVINTERVAL)).intValue();
		
		if (result<1 || result>60) {
			logger.warning("Configuration entry '"+
					CONFIGNAME_ADVINTERVAL+"' must "+
					"be between 1 and 60. Assuming default interval "+
					DEFAULT_ADVINTERVAL+".");
			
			config.saveCorrectedValue(CONFIGNAME_ADVINTERVAL,
					Integer.class,
					DEFAULT_ADVINTERVAL);
			
			return DEFAULT_ADVINTERVAL.intValue();
		}
		
		return result;
	}
	
	/**
	 * Returns the advertisment UDP timeout.
	 * 
	 * @return timeout.
	 */
	public static int getAdvertisementUDPTimeout() {
		int result=((Integer)config.getValue(
				CONFIGNAME_ADVUDPTIMEOUT,
				Integer.class,
				DEFAULT_ADVUDPTIMEOUT)).intValue();
		
		if (result<1 || result>5) {
			logger.warning("Configuration entry '"+
					CONFIGNAME_ADVUDPTIMEOUT+"' must "+
					"be between 1 and 5. Assuming default timeout "+
					DEFAULT_ADVUDPTIMEOUT+".");
			
			config.saveCorrectedValue(CONFIGNAME_ADVUDPTIMEOUT,
					Integer.class,
					DEFAULT_ADVUDPTIMEOUT);
			
			return DEFAULT_ADVUDPTIMEOUT.intValue();
		}
		
		return result;
	}
	
	/**
	 * Returns the advertisment TTL.
	 * 
	 * @return TTL.
	 */
	public static int getAdvertisementTTL() {
		int result=((Integer)config.getValue(
				CONFIGNAME_ADVTTL,
				Integer.class,
				DEFAULT_ADVTTL)).intValue();
		
		if (result<1 || result>64) {
			logger.warning("Configuration entry '"+
					CONFIGNAME_ADVTTL+"' must "+
					"be between 1 and 64. Assuming default ttl "+
					DEFAULT_ADVTTL+".");
			
			config.saveCorrectedValue(CONFIGNAME_ADVTTL,
					Integer.class,
					DEFAULT_ADVTTL);
			
			return DEFAULT_ADVTTL.intValue();
		}
		
		return result;
	}
	
	/**
	 * Returns the IP of this host.
	 * 
	 * @return IP.
	 */
	public static String getIP() {
		String result;
		try {
			result=(String)config.getValue(
					CONFIGNAME_LOCALIP,
					String.class);
			if (result==null) {
				return InetAddress.getLocalHost().getHostAddress();
			}
			result=InetAddress.getByName(result).getHostAddress();
		} catch (UnknownHostException e) {
			// if now ip address could be found then we are not connected
			// anyway ... so let's take the localhost
			config.saveCorrectedValue(CONFIGNAME_LOCALIP,
					String.class,
					result="127.0.0.1");
			
			logger.warning("Configuration entry '"+CONFIGNAME_LOCALIP+"' is "+
					"not a valid IP. Assuming default IP "+result+".");
		}		
		return result;
	}
	
	/**
	 * Returns the UUID of this host.
	 * 
	 * @return Host UUID String.
	 */
	public static String getHostUUID() {
		
		if (hostUUID!=null) {
			return hostUUID;
		}
		
		File configFile=new File(config.getConfigurationFileName());
		File configDir=configFile.getParentFile();
		
		File uuidFile=new File(configDir,"host.uuid");
		
		FileInputStream fis=null;
		try {
			fis=new FileInputStream(uuidFile);
			BufferedReader reader=
				new BufferedReader(new InputStreamReader(fis));
			String input=reader.readLine();
			if (input!=null) {
				hostUUID=input;
				return hostUUID;
			}
		} catch (FileNotFoundException e) {
		} catch (IOException e) {
			logger.warning("Cannot read host UUID file '"+
					uuidFile.getAbsolutePath()+"': "+e.getMessage());
		} finally {
			if (fis!=null) {
				try {
					fis.close();
				} catch (IOException e) {
				}
			}
		}

		hostUUID=new UUID().toString();
		try {
			if (!uuidFile.exists()) {
				uuidFile.createNewFile();
			}
			FileOutputStream fos=new FileOutputStream(uuidFile);
			OutputStreamWriter writer=new OutputStreamWriter(fos);
			
			writer.write(hostUUID);
			
			writer.close();
			fos.close();
		} catch (IOException e) {
			logger.error("Cannot save host UUID to file '"+
					uuidFile.getAbsolutePath()+"': "+e.getMessage());
		}
		
		return hostUUID;
	}
}
