/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.discovery;

import java.util.HashSet;

/**
 * AServiceQuery for service DB.
 * 
 * @author Lukasz Juszczyk
 *
 */
public abstract class AServiceQuery {
	
	protected boolean skipDuplicates=true;
	
	protected HashSet foundServices=new HashSet();
	
	/**
	 * Constructor.
	 *
	 */
	protected AServiceQuery() {
		super();
	}

	/**
	 * Checks if a ServiceInfo matches and whether it has changed since
	 * the last check.
	 * 
	 * @param info ServiceInfo instance.
	 * @return <code>true</code> if matches.
	 */
	private boolean matchesWithoutDuplicateNotification(
			ServiceInfo info) {

		if ( !matches(info) || foundServices.contains(info) ) {
			return false; 
		}

		foundServices.add(info);
		
		return true;
	}
	
	/**
	 * Checks whether the subscribed should be notified.
	 * 
	 * @param service ServiceInfo instance.
	 * @return <code>true</code> if matches.
	 */
	protected final boolean check(ServiceInfo service) {
		if (noDuplicates()) {
			if (!matchesWithoutDuplicateNotification(service)) {
				return false;
			}
		} else {
			if (!matches(service)) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * Checks if a ServiceInfo matches.
	 * 
	 * @param info ServiceInfo instance.
	 * @return <code>true</code> if matches.
	 */
	protected abstract boolean matches(ServiceInfo info);

	/**
	 * Checks whether this query should skip duplicate notifications
	 * about the same services. 
	 * 
	 * @return <code>true</code> if duplicates should be skipped, otherwise
	 * <code>false</code>.
	 */
	public boolean noDuplicates() {
		return skipDuplicates;
	}

	/**
	 * Sets whether this query should skip duplicate notifications
	 * about the same services.
	 * 
	 * @param skipDuplicates Should be <code>true</code> if duplicates 
	 * should be skipped, otherwise should be <code>false</code>. 
	 */
	public void setNoDuplicates(boolean skipDuplicates) {
		this.skipDuplicates = skipDuplicates;
	}

	/**
	 * Declares a service as "unseen" which means, that notifications about
	 * if will be allowed again.
	 * 
	 * @param info Service info.
	 */
	public void declareAsUnseen(ServiceInfo info) {
		foundServices.remove(info);
	}
}