/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.discovery;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.Socket;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import org.ksoap2.serialization.KvmSerializable;
import org.ksoap2.serialization.SoapObject;
import org.xmlpull.v1.XmlPullParserException;

import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.MiddlewareConfiguration;
import at.ac.tuwien.vitalab.middleware.soap.LocalChannel;
import at.ac.tuwien.vitalab.middleware.soap.ServiceContainer;
//import at.ac.tuwien.vitalab.middleware.soap.WSClassLoader;
import at.ac.tuwien.vitalab.middleware.soap.header.SoapHeaders;
import at.ac.tuwien.vitalab.middleware.soap.header.RecipientIDHeader;
import at.ac.tuwien.vitalab.middleware.soap.services.InfoService;
//import at.ac.tuwien.vitalab.middleware.util.Base64;
import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;
import edu.wustl.mobilab.sliver.soap.Channel;
import edu.wustl.mobilab.sliver.soap.MustUnderstandException;
import edu.wustl.mobilab.sliver.soap.j2se.HTTPChannel;
import edu.wustl.mobilab.sliver.soap.j2se.SocketChannel;

/**
 * Information about the services in the network. Also contains 
 * functionality to invoke them.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class ServiceInfo {
	
	private static Logger logger=LoggerFactory.getLogger(ServiceInfo.class);
	
	// predefined attributes
	public static final String ID="id";
	public static final String NAMESPACE="namespace";
	public static final String URL="url";
	public static final String HOSTID="hostid";
	
	// pointer to infoservice of the host
	private ServiceInfo infoService=null;

	// pointer to the localInfoService if one exists
	private static ServiceInfo localInfoService=null;
	
	// if this service is local, this is a pointer to its implementation
	private AWebService localService=null;
	
	// hashset containing the collections of retrieved data so that we
	// know which ones don't need to be retrieved again
	private final HashSet retrievedData=new HashSet();
	
	private final static String RETR_ATTS="attributes";
	private final static String RETR_OPS="operationNames";
	
	// service information
	private final Hashtable attributes=new Hashtable();
	private final Vector operationNames=new Vector();
	private final Hashtable returnTypes=new Hashtable();
	private final Hashtable paramTypes=new Hashtable();
	private final Hashtable descriptions=new Hashtable();
	
	private long lastServiceRefreshTimestamp=0;
	
	// to see wether the service has not been updated during the last interval
	// this means, it is inactive
	private int interval=MiddlewareConfiguration.getAdvertisementInterval();
	
	// service database where this service info is located
	private ServiceDB serviceDB;

	// for updates about service availability
	private boolean lastUpdateAsAvailable=true;
	private HashSet updateReceivers=new HashSet();
	
	/**
	 * Constructor which should be used to create an entry of a found 
	 * InfoService.
	 * 
	 * @param url URL of the info service.
	 * @param uuid UUID of its host.
	 * @param serviceDB Local service database.
	 */
	protected ServiceInfo(String url, String uuid, ServiceDB serviceDB) {
		
		this.serviceDB=serviceDB;
		this.infoService=this;
		this.attributes.put(URL, url);
		this.attributes.put(HOSTID, uuid);
		this.attributes.put(ID, InfoService.NAME);
		this.attributes.put(NAMESPACE, InfoService.NAMESPACE);
		refresh();
	}

	/**
	 * Constructor which should be used to create an entry of usual service.

	 * @param infoService Reference to its info service.
	 * @param namespace Its namespace.
	 * @param uuid UUID of its host.
	 * @param serviceDB Local service database.
	 */
	protected ServiceInfo(ServiceInfo infoService, 
			String namespace, String uuid, ServiceDB serviceDB) {
		
		this.serviceDB=serviceDB;
		this.infoService=infoService;		
		this.attributes.put(HOSTID, uuid);
		this.attributes.put(NAMESPACE, namespace);
		refresh();		
	}
	
	/**
	 * Constructor which should be used to create an entry for a local service.
	 * 
	 * @param localWS Local Web service.
	 * @param serviceDB Local service database.
	 */
	protected ServiceInfo(AWebService localWS, ServiceDB serviceDB) {
		
		this.serviceDB=serviceDB;
		this.localService=localWS;		
		this.attributes.putAll(AttributeRetriever.getAttributes(localWS));
		retrievedData.add(RETR_ATTS);
		this.operationNames.addAll(localWS.getOperationNames());
		this.operationNames.addAll(localWS.getOperationNameMapping().keySet());
		retrievedData.add(RETR_OPS);
		
		if (localWS instanceof InfoService) {
			localInfoService=this;
		}
		
		if (localInfoService==null) {
			throw new IllegalStateException("No (local) InfoService found.");
		}
		
		infoService=localInfoService;

	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object arg0) {
		try {
			ServiceInfo argInfo=(ServiceInfo)arg0;
			// it is only equal if hostID and serviceID match
			return argInfo.getHostId().equals(getHostId()) &&
			argInfo.getId().equals(getId());
		} catch (ClassCastException cce) {
			return false;
		} catch (MissingServiceAttributeException e) {
			return false; // should not happen, will be checked before
		}
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		try {			
			return (getId()+" "+getHostId()).hashCode();			
		} catch (MissingServiceAttributeException e) {
			return -1; // should not happen, will be checked before
		}
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String result="";
		try {
			result=getId()+"@"+getUrl()+"("+getHostId()+")";
		} catch (Exception e) {
			result=super.toString();
		}
		return result;
	}

	/**
	 * Checks whether this service is inactive, or whether it was
	 * refreshed.
	 * 
	 * @return <code>true</code> if inactive.
	 */
	public boolean isInactive() {
		
		if (localService!=null) {
			return !serviceDB.hasLocalService(localService);
		}
		
		return ((System.currentTimeMillis()-lastServiceRefreshTimestamp)
				>(interval*2500));			
	}
	
	/**
	 * Checks whether the service is a local one.
	 * 
	 * @return <code>true</code> if local, otherwise <code>false</code>.
	 */
	protected boolean isLocal() {
		return localService!=null;
	}
	
	/**
	 * Refreshes the timestamp.
	 *
	 */
	protected void refresh() {
		lastServiceRefreshTimestamp=System.currentTimeMillis();
		
		checkAvailabilityUpdates();
	}
	
	/**
	 * Declares this service as inactive.
	 *
	 */
	protected void setInactive() {
		lastServiceRefreshTimestamp=0;
		
		checkAvailabilityUpdates();
	}
	
	/**
	 * Checks if the service is active, otherwise throws an exception.
	 * 
	 * @throws InactiveServiceException Exception.
	 */
	private void checkActivity() throws InactiveServiceException {
		if (isInactive()) {
			try {
				throw new InactiveServiceException(
						"Service '"+getId()+"' is inactive.");
			} catch (MissingServiceAttributeException e) {
				throw new InactiveServiceException(
						"Service is inactive.");
			}
		}
	}
	
	/**
	 * Retrieves the service attributes from the network.
	 */
	private void retrieveAttributes() {

		synchronized (attributes) {
			if (retrievedData.contains(RETR_ATTS)) {
				return;
			}
			
			Hashtable atts=null;
			
			//first try udp
			try {
				atts=ServiceRegistry.retrieveAttributesViaUDP(
						infoService,getNamespace());
			} catch (Exception e) {
				logger.debug("Cannot retrieve attributes via udp: "+
						e.getMessage());

				// alternatively try a soap call				
				try {
					atts=(Hashtable)infoService.invokeOperation(
							"getServiceAttributes",
							new String[]{getNamespace()}, 
							null);
				} catch (Exception e1) {
					logger.debug("Cannot retrieve attributes: "+
							e.getMessage());
					return;
				}				

			}
			
			attributes.putAll(atts);
			retrievedData.add(RETR_ATTS);
			
			// notify subscribers
			serviceDB.checkServiceForSubscriptions(this);
		}		
	}
	
	/**
	 * Returns all attributes.
	 * 
	 * @return Hashtable with attributes.
	 */
	public Hashtable getAttributes() {
		retrieveAttributes();
		return new Hashtable(attributes);
	}
	
	/**
	 * Returns an attriubute.
	 * 
	 * @param key Attribute key.
	 * @return Attribute value or <code>null</code> if nothing was found.
	 */
	public String getAttribute(String key) {
		if (!attributes.containsKey(key)) {
			retrieveAttributes();		
		}
		return (String)attributes.get(key);
	}

	/**
	 * Returns the namespace attribute.
	 * 
	 * @return Namespace.
	 * @throws MissingServiceAttributeException If the attribute has not
	 * been published. 
	 */
	public String getNamespace() throws MissingServiceAttributeException {
		return getAttributeValue(NAMESPACE);
	}

	/**
	 * Returns the URL attribute.
	 * 
	 * @return URL.
	 * @throws MissingServiceAttributeException If the attribute has not
	 * been published. 
	 */
	public String getUrl() throws MissingServiceAttributeException {
		
		if (this.isLocal()) {
			try {
				ServiceContainer local=
					ServiceContainer.getLocalServiceContainer();
				
				return local.getServiceURL();
			} catch (Exception e) {
				// return what was here before...
			}
		} 
		return getAttributeValue(URL);
	}

	/**
	 * Returns the host of the service info, retrieved from the URL attribute.
	 * 
	 * @return Host String.
	 * @throws MissingServiceAttributeException If the URL attribute has not
	 * been published. 
	 * @throws MalformedURLException If the URL is malformed.
	 */
	public String getHost() throws MissingServiceAttributeException, 
	MalformedURLException {
		String url=getUrl();
		int endOfProtocol=url.indexOf("://");
		if (endOfProtocol!=-1) {
			url="http"+url.substring(endOfProtocol);
		}
		return new URL(url).getHost();
	}

	/**
	 * Returns the SOAP port of the service info, 
	 * retrieved from the URL attribute.
	 * 
	 * @return Port number.
	 * @throws MissingServiceAttributeException If the URL attribute has not
	 * been published. 
	 * @throws MalformedURLException If the URL is malformed.
	 */
	public int getSOAPPort() throws MissingServiceAttributeException, 
	MalformedURLException {
		String url=getUrl();
		int endOfProtocol=url.indexOf("://");
		if (endOfProtocol!=-1) {
			url="http"+url.substring(endOfProtocol);
		}
		return new URL(url).getPort();
	}
	
	/**
	 * Returns the UDP command port of the service info, 
	 * retrieved from the URL attribute.
	 * 
	 * @return Port number.
	 * @throws MissingServiceAttributeException If the URL attribute has not
	 * been published. 
	 * @throws MalformedURLException If the URL is malformed.
	 */
	public int getUDPCommandPort() throws MissingServiceAttributeException, 
	MalformedURLException {
		String url=getUrl();
		int endOfProtocol=url.indexOf("://");
		if (endOfProtocol!=-1) {
			url="http"+url.substring(endOfProtocol);
		}
		return new URL(url).getPort()-1;
	}
	
	/**
	 * Updates the URL. Necessary when the location of the host changes.
	 * 
	 * @param url URL.
	 */
	protected void updateUrl(String url) {
		attributes.put(URL, url);
	}
	
	/**
	 * Returns the ID attribute.
	 * 
	 * @return ID.
	 * @throws MissingServiceAttributeException If the attribute has not
	 * been published. 
	 */
	public String getId() throws MissingServiceAttributeException {
		return getAttributeValue(ID);
	}

	/**
	 * Returns the Host ID attribute.
	 * 
	 * @return Host ID.
	 * @throws MissingServiceAttributeException If the attribute has not
	 * been published. 
	 */
	public String getHostId() throws MissingServiceAttributeException {
		String result=getAttributeValue(HOSTID);
		if (result!=null) {
			return result.replace(':', '.');
		} else {
			return null;
		}
	}
	
	/**
	 * Returns the number of hops between the host of the service and
	 * the local host.
	 * 
	 * @return Number of hops. 0 means local service. -1 means unavailable.
	 * -2 means that the hop count is unknown.
	 */
	public int getHopsToHost() {
		if (this.isInactive()) {
			return -1;
		}
		if (this.isLocal()) {
			return 0;
		}
		try {
			return serviceDB.getHopsToHost(getHostId());			
		} catch (Exception iae) {
			return -2;
		}
	}
	
	/**
	 * Returns the value of an attribute.
	 * 
	 * @param attribName Attribute name.
	 * @return Value.
	 * @throws MissingServiceAttributeException If the attribute was not found.
	 */
	private String getAttributeValue(String attribName) 
	throws MissingServiceAttributeException {
		String result=getAttribute(attribName);
		if (result==null) {
			throw new MissingServiceAttributeException(
					"Attribute '"+attribName+"' unknown.");
		}
		return result;
	}
	
	/**
	 * Returns a transport channel to the service.
	 * 
	 * @return Transport channel.
	 * @throws IOException If IO errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Channel getChannel() throws IOException, 
	InactiveServiceException, MissingServiceAttributeException {
		
		checkActivity();	
		
		String url=getUrl();
		
		Channel channel=null;
		
		if (this.isLocal()) {
			channel=new LocalChannel();
		} else {

			if (url.startsWith("socket://")) {
				
				// dirty little trick to get rid of the exception due
				// to an unknown protocol...
				URL tmpURL=new URL("http"+url.substring(6));
				
				Socket socket=new Socket(tmpURL.getHost(),
						tmpURL.getPort());	
				
				channel=new SocketChannel(socket);	
			} else {
				if (!url.startsWith("http://")) {
					logger.warning("Unknown protocol '"+url+"'. " +
							"Assuming 'http'.");
				}
				
				HttpURLConnection connection=
					(HttpURLConnection)new URL(url+"/").openConnection();
				
				channel=new HTTPChannel(connection,getNamespace());
			}
		}
		
		RecipientIDHeader toHeader=new RecipientIDHeader(this.getHostId());
		channel.getOutgoingHeaderElements().addHeader(toHeader);
		
		return channel;
	}
	
	/**
	 * Retrieves all information about this service.
	 * 
	 * @throws UnknownHostException If the service's host is invalid.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
 	 * @throws ClassCastException If the service responds with an
	 * invalid object.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published.
	 * @throws ClassNotFoundException If a referenced type is invalid/unknown.
	 */
	private void retrieveAllInfo() throws UnknownHostException, IOException, 
	XmlPullParserException, MustUnderstandException, ClassCastException, 
	InactiveServiceException, MissingServiceAttributeException, 
	ClassNotFoundException {
		
		synchronized (operationNames) {
			synchronized (returnTypes) {
				synchronized (paramTypes) {				
					synchronized (attributes) {
						synchronized (descriptions) {
							_retrieve();
						}
					}
				}
			}
		}
	}
	
	/**
	 * Really retrieves all information about this service.
	 * 
	 * @throws UnknownHostException If the service's host is invalid.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
 	 * @throws ClassCastException If the service responds with an
	 * invalid object.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published.
	 * @throws ClassNotFoundException If a referenced type is invalid/unknown.
	 */
	private void _retrieve() throws UnknownHostException, IOException, 
	XmlPullParserException, MustUnderstandException, ClassCastException, 
	InactiveServiceException, MissingServiceAttributeException, 
	ClassNotFoundException {
		
		Hashtable result=
			(Hashtable)infoService.invokeOperation(
					"getServiceInfo", 
					new String[]{getNamespace()}, 
					null);
		
		Hashtable atts=(Hashtable)result.get("attributes");
		
		if (atts!=null && !retrievedData.contains(RETR_ATTS)) {
			attributes.putAll(atts);							
			retrievedData.add(RETR_ATTS);
			
			// notify subscribers
			serviceDB.checkServiceForSubscriptions(this);
		} 
		
		Enumeration keyEnum=result.keys();
		
		Vector opNames=new Vector();
		
		while (keyEnum.hasMoreElements()) {
			String key=keyEnum.nextElement().toString();
			if (key.startsWith("op.")) {
				
				String operationName=key.substring(3);
				opNames.add(operationName);
				
				// return type
				try {
					String returnTypeResp=
						((Hashtable)result.get(key)).get(
								"return").toString();
					
					returnTypes.put(operationName, returnTypeResp);
					
				} catch (Exception e) {
					// ignore
				}				
				
				// param types
				try {
					Vector paramTypesResp=
						(Vector)((Hashtable)result.get(key)).get("params");
					
					paramTypes.put(operationName, paramTypesResp);
					
				} catch (Exception e) {
					// ignore
				}
				
				//descriptions
				try {
					Hashtable descResp=(Hashtable)((Hashtable)result.get(
							key)).get("description");
					if (descResp==null) {
						descResp=new Hashtable();
					}
					descriptions.put(operationName, descResp);
					
				} catch (Exception e) {
					// ignore
				}
			}
		}
		
		if (!retrievedData.contains(RETR_OPS)) {
			operationNames.addAll(opNames);
			retrievedData.add(RETR_OPS);
		}
	}
	
	/**
	 * Returns the names of all operations of this Web service.
	 * 
	 * @return List of names.
	 * @throws UnknownHostException If the service's host is invalid.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws ClassCastException If the service responds with an
	 * invalid object.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Vector getOperationnames() throws UnknownHostException, IOException, 
	XmlPullParserException, MustUnderstandException, ClassCastException, 
	InactiveServiceException, MissingServiceAttributeException {
		
		synchronized (operationNames) {
			if (retrievedData.contains(RETR_OPS)) {
				return operationNames;
			}
			
			try {
				retrieveAllInfo();				
			} catch (Exception e) {
				e.printStackTrace();
				logger.debug("Cannot retrieve service info: "+e.getMessage());
			}

			if (retrievedData.contains(operationNames)) {
				return operationNames;
			}

			Vector result=(Vector)infoService.invokeOperation(
					"getServiceOperations", 
					new String[] {getNamespace()}, 
					null);
			
			operationNames.addAll(result);
			
			retrievedData.add(RETR_OPS);
			
			return operationNames;	
		}
	}
	
	/**
	 * Returns the return type of this Web service's operation.
	 * 
	 * @param operationName Name of the operation.
	 * @return Return type name.
	 * @throws UnknownHostException If the service's host is invalid.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws ClassNotFoundException If the return type is invalid/unknown.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public String getReturnType(String operationName) 
	throws UnknownHostException, IOException, XmlPullParserException, 
	MustUnderstandException, ClassNotFoundException, 
	InactiveServiceException, MissingServiceAttributeException {
		
		synchronized (returnTypes) {
			if (returnTypes.containsKey(operationName)) {
				return (String)returnTypes.get(operationName);
			}
			
			try {
				retrieveAllInfo();				
			} catch (Exception e) {
				logger.debug("Cannot retrieve service info: "+e.getMessage());
			}
			
			if (returnTypes.containsKey(operationName)) {
				return (String)returnTypes.get(operationName);
			}

			Object response=infoService.invokeOperation(
					"getServiceOperationReturnType", 
					new String[]{getNamespace(), operationName}, null);
			
			String result=response.toString();
			
			returnTypes.put(operationName, result);
			
			return result;			
		}
	}
	
	/**
	 * Returns the parameter types of this Web service's operation.
	 * 
	 * @param operationName Name of the operation.
	 * @return Vector of parameter type names.
	 * @throws UnknownHostException If the service's host is invalid.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws ClassNotFoundException If at least one of the parameter 
	 * types is invalid/unknown.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Vector getParameterTypes(String operationName) 
	throws UnknownHostException, IOException, XmlPullParserException, 
	MustUnderstandException, ClassNotFoundException, 
	InactiveServiceException, MissingServiceAttributeException {

		synchronized (paramTypes) {

			if (paramTypes.containsKey(operationName)) {
				return (Vector)paramTypes.get(operationName);
			}
			
			try {
				retrieveAllInfo();				
			} catch (Exception e) {
				logger.debug("Cannot retrieve service info: "+e.getMessage());
			}

			if (paramTypes.containsKey(operationName)) {
				return (Vector)paramTypes.get(operationName);
			}

			Vector response=(Vector)infoService.invokeOperation(
					"getServiceOperationParameterTypes", 
					new String[]{getNamespace(), operationName}, 
					null);
			
			Vector result=new Vector();
			
			for (int i=0;i<response.size();i++) {
				result.add(response.get(i).toString());
			}
			
			paramTypes.put(operationName, result);

			return result;
		}
	}
	
	/**
	 * Returns the description attributes of this Web service's operation.
	 * 
	 * @param operationName Name of the operation.
	 * @return Hashtable of description attributes.
	 * @throws UnknownHostException If the service's host is invalid.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Hashtable getDescription(String operationName) 
	throws UnknownHostException, IOException, XmlPullParserException, 
	MustUnderstandException,  
	InactiveServiceException, MissingServiceAttributeException {

		synchronized (descriptions) {

			if (descriptions.containsKey(operationName)) {
				return (Hashtable)descriptions.get(operationName);
			}
			
			try {
				retrieveAllInfo();				
			} catch (Exception e) {
				logger.debug("Cannot retrieve service info: "+e.getMessage());
			}

			if (descriptions.containsKey(operationName)) {
				return (Hashtable)descriptions.get(operationName);
			}

			Hashtable response=(Hashtable)infoService.invokeOperation(
					"getServiceOperationDescription", 
					new String[]{getNamespace(), operationName}, 
					null);
			
			descriptions.put(operationName, response);
			
			return response;
		}
	}
	
	/**
	 * Transform a class name to the class instance.
	 * 
	 * @param name Class name.
	 * @return Class instance.
	 * @throws ClassNotFoundException If the class name is invalid/unknown.
	 */
	public static Class nameToClass(String name) throws ClassNotFoundException {
		if ("byte".equals(name)) return Byte.class;
		if ("short".equals(name)) return Short.class;
		if ("int".equals(name)) return Integer.class;
		if ("long".equals(name)) return Long.class;
		if ("float".equals(name)) return Float.class;
		if ("double".equals(name)) return Double.class;
		if ("char".equals(name)) return Character.class;
		if ("boolean".equals(name)) return Boolean.class;
		
		return Class.forName(name);			
	}
	
//	/**
//	 * Returns the instance object of a message type.
//	 * 
//	 * @param typeName Type name.
//	 * 
//	 * @return Serializable message type.
//	 * 
//	 * @throws InstantiationException If the type cannot be instiated.
//	 * @throws IllegalAccessException If the type cannot be instiated.
//	 * @throws IOException If IO errors occur.
//	 * @throws XmlPullParserException If the XML messages cannot be 
//	 * parsed properly. 
//	 * @throws MustUnderstandException If the service cannot understand
//	 * the request.
//	 * @throws InactiveServiceException If the service is not active
//	 * anymore.
//	 * @throws MissingServiceAttributeException If some of the 
//	 * necessary service attributes have not been published. 
//	 */
//	public KvmSerializable getMessageType(String typeName) 
//	throws IOException, XmlPullParserException, MustUnderstandException, 
//	InactiveServiceException, MissingServiceAttributeException, 
//	InstantiationException, IllegalAccessException {
//		
//		String response=(String)infoService.invokeOperation(
//				"getSerializedType", 
//				new String[]{typeName}, 
//				null);
//	
//		WSClassLoader wsc=new WSClassLoader(
//				Thread.currentThread().getContextClassLoader());
//		
//		byte[] bytes=Base64.decode(response);
//		
//		Class clazz=wsc.classFromBytes(typeName, bytes);
//		
//		if (!KvmSerializable.class.isAssignableFrom(clazz)) {
//			throw new IOException("Invalid type '"+typeName+"'" +": " +
//					"no KvmSerializable.");
//		}
//		
//		Object o=clazz.newInstance();
//						
//		return (KvmSerializable)o;
//	}
	
	/**
	 * Subscribes to receive updates about the availability of this service.
	 * 
	 * @param rec Update receiver.
	 */
	public void subscribeToAvailabilityUpdates(IServiceUpdateReceiver rec) {
		updateReceivers.add(rec);
	}

	/**
	 * Unsubscribes from updates about the availability of this service.
	 * 
	 * @param rec Update receiver.
	 */
	public void unsubscribeFromAvailabilityUpdates(IServiceUpdateReceiver rec) {
		updateReceivers.remove(rec);
	}

	/**
	 * Checks whether availability updates must be fired. 
	 *
	 */
	protected void checkAvailabilityUpdates() {

		
		synchronized (updateReceivers) {
			
			boolean currentAvailability=!this.isInactive();
			
			if (currentAvailability!=lastUpdateAsAvailable) {
				
				lastUpdateAsAvailable=currentAvailability;
				
				Iterator iter=updateReceivers.iterator();
				
				while (iter.hasNext()) {
					IServiceUpdateReceiver rec=
						(IServiceUpdateReceiver)iter.next();
					
					if (currentAvailability) {
						rec.serviceAvailable(this);
					} else {
						rec.serviceUnavailable(this);
					}
				}
			}	
		}
	}
	
	/**
	 * Invokes the Web service's operation.
	 * 
	 * @param operationName Name of the operation.
	 * @param callbackReceiver Callback receiver for asynchronous invocations.
	 * If <code>null</code> a synchronous invocation will be performed. 
	 * @param parameters Parameters.
	 * @return Return value.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Object invokeOperation(String operationName, Object[] parameters, 
			ICallbackReceiver callbackReceiver)
	throws IOException, XmlPullParserException, 
	MustUnderstandException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		return invokeOperation(
				operationName, parameters, callbackReceiver, null);
	}
	
	/**
	 * Invokes the Web service's operation.
	 * 
	 * @param operationName Name of the operation.
	 * @param callbackReceiver Callback receiver for asynchronous invocations.
	 * If <code>null</code> a synchronous invocation will be performed. 
	 * @param parameters Parameters.
	 * @param headers Soap headers. Can be <code>null</code> in order to 
	 * be ignored.
	 * @return Return value.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Object invokeOperation(String operationName, Object[] parameters, 
			ICallbackReceiver callbackReceiver, SoapHeaders headers) 
	throws IOException, XmlPullParserException, 
	MustUnderstandException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		SoapObject soapObject=new SoapObject(getNamespace(),operationName);
		for (int i=0;i<parameters.length;i++) {
			soapObject.addProperty("arg"+i, parameters[i]);
		}
		
		// invocation
		return invokeOperation(soapObject, callbackReceiver, headers);		
	}
	
	/**
	 * Invokes the Web service's operation.
	 * 
	 * @param operationName Name of the operation.
	 * @param callbackReceiver Callback receiver for asynchronous invocations.
	 * If <code>null</code> a synchronous invocation will be performed. 
	 * @param parameter Parameter.
	 * @return Return value.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Object invokeOperation(String operationName, 
			KvmSerializable parameter, ICallbackReceiver callbackReceiver) 
	throws IOException, XmlPullParserException, 
	MustUnderstandException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		return invokeOperation(
				operationName, parameter, callbackReceiver, null);
	}
	
	/**
	 * Invokes the Web service's operation.
	 * 
	 * @param operationName Name of the operation.
	 * @param callbackReceiver Callback receiver for asynchronous invocations.
	 * If <code>null</code> a synchronous invocation will be performed. 
	 * @param parameter Parameter.
	 * @param headers Soap headers. Can be <code>null</code> in order to 
	 * be ignored.
	 * @return Return value.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Object invokeOperation(String operationName, 
			KvmSerializable parameter, ICallbackReceiver callbackReceiver,
			SoapHeaders headers) 
	throws IOException, XmlPullParserException, 
	MustUnderstandException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		SoapObject soapObject=new SoapObject(getNamespace(),operationName);
		for (int i=0;i<parameter.getPropertyCount();i++) {
			soapObject.addProperty("arg"+i, parameter.getProperty(i));
		}
		
		// invocation
		return invokeOperation(soapObject, callbackReceiver, headers);
	}
	
	/**
	 * Invokes the Web service's operation.
	 * 
	 * @param soapObject Message.
	 * @param callbackReceiver Callback receiver for asynchronous invocations.
	 * If <code>null</code> a synchronous invocation will be performed. 
	 * @return Return value or messageID in case of asynchronous invocation.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Object invokeOperation(SoapObject soapObject, 
			ICallbackReceiver callbackReceiver) 
	throws IOException, XmlPullParserException, 
	MustUnderstandException, InactiveServiceException, 
	MissingServiceAttributeException {
		return invokeOperation(soapObject, callbackReceiver, null);
	}
	
	/**
	 * Invokes the Web service's operation.
	 * 
	 * @param soapObject Message.
	 * @param callbackReceiver Callback receiver for asynchronous invocations.
	 * If <code>null</code> a synchronous invocation will be performed. 
	 * @param headers Soap headers. Can be <code>null</code> in order to 
	 * be ignored.
	 * @return Return value or messageID in case of asynchronous invocation.
	 * @throws IOException If IO errors occur.
	 * @throws XmlPullParserException If the XML messages cannot be 
	 * parsed properly. 
	 * @throws MustUnderstandException If the service cannot understand
	 * the request.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public Object invokeOperation(SoapObject soapObject, 
			ICallbackReceiver callbackReceiver, SoapHeaders headers) 
	throws IOException, XmlPullParserException, 
	MustUnderstandException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		Channel channel=null;
		try {

			channel=getChannel();

			channel.getOutgoingHeaderElements().addHeaders(headers);
			
			Object result=null;
			
			if (callbackReceiver!=null) {
				// async invocation
				result=channel.callAsync(soapObject,callbackReceiver);								
			} else {
				result=channel.call(soapObject);		// invocation
			}
			
			refresh();
			
			return result;
			
		} catch (IOException ioe) {
			setInactive();
			throw ioe;
		} finally {
			if (channel!=null) {
				try {
					channel.close();					
				} catch (IOException ioe) {
					// ignore
				}
			}
		}
	}

	/**
	 * Sends a message to the Web service.
	 * 
	 * @param messageName Message name.
	 * @param parameters Parameters.
	 * @throws IOException If IO errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public void sendMessage(String messageName, Object[] parameters) 
	throws IOException, InactiveServiceException, 
	MissingServiceAttributeException {
		sendMessage(messageName, parameters, null);
	}
	
	/**
	 * Sends a message to the Web service.
	 * 
	 * @param messageName Message name.
	 * @param parameters Parameters.
	 * @param headers Soap headers. Can be <code>null</code> in order to 
	 * be ignored.
	 * @throws IOException If IO errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public void sendMessage(String messageName, Object[] parameters,
			SoapHeaders headers) 
	throws IOException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		SoapObject soapObject=new SoapObject(getNamespace(),messageName);
		for (int i=0;i<parameters.length;i++) {
			soapObject.addProperty("arg"+i, parameters[i]);
		}
	
		sendMessage_(soapObject, headers);
	}
	
	/**
	 * Sends a message to the Web service.
	 * 
	 * @param messageName Message name.
	 * @param message Message.
	 * @throws IOException If IO errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public void sendMessage(String messageName, KvmSerializable message) 
	throws IOException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		sendMessage(messageName, message, null);
	}
	
	/**
	 * Sends a message to the Web service.
	 * 
	 * @param messageName Message name.
	 * @param message Message.
	 * @param headers Soap headers. Can be <code>null</code> in order to 
	 * be ignored.
	 * @throws IOException If IO errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public void sendMessage(String messageName, KvmSerializable message,
			SoapHeaders headers) 
	throws IOException, InactiveServiceException, 
	MissingServiceAttributeException {
		
		SoapObject soapObject=new SoapObject(getNamespace(),messageName);
		for (int i=0;i<message.getPropertyCount();i++) {
			soapObject.addProperty("arg"+i, message.getProperty(i));
		}
	
		sendMessage_(soapObject, headers);
	}

	/**
	 * Sends a message to the Web service.
	 * 
	 * @param soapObject Message.
	 * @throws IOException If IO errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public void sendMessage(SoapObject soapObject) 
	throws IOException, InactiveServiceException, 
	MissingServiceAttributeException {

		sendMessage_(soapObject, null);
	}

	/**
	 * Sends a message to the Web service.
	 * 
	 * @param message Message.
	 * @param headers Soap headers. Can be <code>null</code> in order to 
	 * be ignored.
	 * @throws IOException If IO errors occur.
	 * @throws InactiveServiceException If the service is not active
	 * anymore.
	 * @throws MissingServiceAttributeException If some of the 
	 * necessary service attributes have not been published. 
	 */
	public void sendMessage_(SoapObject message, SoapHeaders headers) 
	throws InactiveServiceException, 
	MissingServiceAttributeException, IOException {
		
		Channel channel=null;
		try {
			
			channel=getChannel();
			
			channel.getOutgoingHeaderElements().addHeaders(headers);
			
			channel.sendObject(message);
			
			refresh();
			
		} catch (IOException ioe) {
			setInactive();
			throw ioe;
		} finally {
			if (channel!=null) {
				try {
					channel.close();					
				} catch (IOException ioe) {
					// ignore
				}
			}
		}
	}
}
