/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap;

import java.io.IOException;

import org.xmlpull.v1.XmlPullParserException;

import at.ac.tuwien.vitalab.middleware.Middleware;
import at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery;
import at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber;
import at.ac.tuwien.vitalab.middleware.discovery.InactiveServiceException;
import at.ac.tuwien.vitalab.middleware.discovery.MissingServiceAttributeException;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceDB;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo;
import at.ac.tuwien.vitalab.middleware.soap.header.MessageIDHeader;
import at.ac.tuwien.vitalab.middleware.soap.header.RecipientIDHeader;
import at.ac.tuwien.vitalab.middleware.soap.header.RelatesToHeader;
import at.ac.tuwien.vitalab.middleware.soap.header.AsyncCallHeader;
import at.ac.tuwien.vitalab.middleware.soap.header.SoapHeaders;
import at.ac.tuwien.vitalab.middleware.soap.services.CallbackService;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;
import edu.wustl.mobilab.sliver.soap.Channel;
import edu.wustl.mobilab.sliver.soap.MustUnderstandException;

/**
 * Channel for asynchronous responses.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class AsyncReplyChannel extends Channel implements IServiceSubscriber {

	protected AsyncCallHeader asyncCallHeader;
	protected MessageIDHeader messageIDHeader;
	protected ServiceDB serviceDB;
	
	private static Logger logger=
		LoggerFactory.getLogger(AsyncReplyChannel.class);
	
	/**
	 * Constructor.
	 * 
	 * @param asyncCallHeader ReplyTo header of request message.
	 * @param messageIDHeader MessageID header of request message.
	 */
	public AsyncReplyChannel(AsyncCallHeader asyncCallHeader, 
			MessageIDHeader messageIDHeader) {
		super();
		this.asyncCallHeader = asyncCallHeader;
		this.messageIDHeader = messageIDHeader;
		serviceDB=Middleware.getServiceDatabase();
	}

	/* (non-Javadoc)
	 * @see edu.wustl.mobilab.sliver.soap.Channel#closeImpl()
	 */
	protected void closeImpl() throws IOException {
		// nothing to do
	}

	/* (non-Javadoc)
	 * @see edu.wustl.mobilab.sliver.soap.Channel#receiveObject()
	 */
	public Object receiveObject() throws IOException, XmlPullParserException,
			MustUnderstandException {
		// just for sending
		throw new IOException("Receiving objects via this channel " +
		"is not supported.");
	}

	/* (non-Javadoc)
	 * @see edu.wustl.mobilab.sliver.soap.Channel#sendObject(java.lang.Object)
	 */
	public void sendObject(Object object) throws IOException {
		
		CallbackQuery query=new CallbackQuery(asyncCallHeader);
		
		query.setMessage(SoapResponse.retrieveResultObject(object));	
		serviceDB.subscribe(query, this);
	}

	/**
	 * Tries to send the object to the passed service.
	 * 
	 * @param info Service.
	 * @param object Message object.
	 * @return <code>true</code> if the sending was successful, otherwise
	 * <code>false</code>.
	 */
	private boolean sendIt(ServiceInfo info, Object object) {
		try {
			SoapHeaders headers=new SoapHeaders();
			
			headers.addHeader(new RelatesToHeader(messageIDHeader));
			headers.addHeader(new RecipientIDHeader(info.getHostId()));
			
			info.invokeOperation(
					"callback", new Object[]{object}, null, headers);
			
			return true;
			
		} catch (InactiveServiceException e) {
			// wait
		} catch (MissingServiceAttributeException e) {
			// wait
		} catch (XmlPullParserException e) {
			// response should be empty
		} catch (MustUnderstandException e) {
			// real response will come in a seperate message
		} catch (IOException e) {
			// retry next time
			e.printStackTrace();
			logger.debug("Cannot send async reply: "+e.getMessage());
		}
		return false;
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.discovery.IServiceSubscriber#notification(at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery, at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
	 */
	public void notification(AServiceQuery query, ServiceInfo info) {
		if (query instanceof CallbackQuery) {
			CallbackQuery q=(CallbackQuery)query;
			if (q.getMessage()!=null && serviceDB.isSubscribed(query)) { 
				if (sendIt(info,q.getMessage())) {
					serviceDB.unsubscribe(q);						
				} else {
					// give us another notification
					q.declareAsUnseen(info);
				}
			}
		}
	}
	
	/**
	 * AServiceQuery for callback methods.
	 * 
	 * @author Lukasz Juszczyk
	 *
	 */
	private class CallbackQuery extends AServiceQuery {

		private AsyncCallHeader asyncCallHeader;
		private Object message=null;

		/**
		 * Constructor.
		 * 
		 * @param asyncCallHeader Header specifying where the response should
		 * go to.
		 */
		public CallbackQuery(AsyncCallHeader asyncCallHeader) {
			super();
			this.asyncCallHeader=asyncCallHeader;
		}

		/**
		 * Returns the message object.
		 * 
		 * @return Message.
		 */
		public Object getMessage() {
			return message;
		}

		/**
		 * Sets the message object.
		 * 
		 * @param message Message.
		 */
		public void setMessage(Object message) {
			this.message = message;
		}

		/* (non-Javadoc)
		 * @see at.ac.tuwien.vitalab.middleware.discovery.AServiceQuery#matches(at.ac.tuwien.vitalab.middleware.discovery.ServiceInfo)
		 */
		public boolean matches(ServiceInfo info) {
			try {
				return info.getId().equals(
						CallbackService.NAME) &&
				info.getNamespace().equals(
						CallbackService.NAMESPACE) &&				
				info.getHostId().equals(
						asyncCallHeader.getHostID());
			} catch (Exception e) {
				return false;
			}
		}
		
	}
}
