/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap;

import java.io.IOException;

import at.ac.tuwien.vitalab.middleware.MiddlewareConfiguration;
import at.ac.tuwien.vitalab.middleware.AWebService;

import edu.wustl.mobilab.sliver.soap.SOAPServer;
import edu.wustl.mobilab.sliver.soap.Transport;
import edu.wustl.mobilab.sliver.soap.j2se.JettyTransport;
import edu.wustl.mobilab.sliver.soap.j2se.SocketTransport;

/**
 * SOAP Server.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class ServiceContainer {

	private String serviceURL=null;
	private SOAPServer server=null;
	
	private static ServiceContainer localContainer=null;
	
	/**
	 * Constructor. Creates a SOAPServer with the passed transport type at the 
	 * desired port.
	 * 
	 * @throws Exception If errors occur while creating the Transport.
	 */
	protected ServiceContainer() throws Exception {
		
		int confPort=MiddlewareConfiguration.getSoapPort();
		String confTransport=MiddlewareConfiguration.getSoapTransportType();
		String allIP="0.0.0.0";
		
		Transport transport=null;
		
		if (confTransport.toLowerCase().equals(
				MiddlewareConfiguration.TRANSPORT_SOCKET)) {
			transport=new SocketTransport(allIP,confPort);
			
		} else  {
			
			String confServicePath=MiddlewareConfiguration.getHttpServicePath();
			
			transport=new JettyTransport(allIP,confPort,confServicePath);
		}
		
		server=new SOAPServer(transport);		
	}
	
	/**
	 * Returns the URL at which all services will are registered.
	 * 
	 * @return URL as String.
	 */
	public String getServiceURL() {

		String myIP=MiddlewareConfiguration.getIP();
		String confTransport=MiddlewareConfiguration.getSoapTransportType();
		int confPort=MiddlewareConfiguration.getSoapPort();

		if (confTransport.toLowerCase().equals(
				MiddlewareConfiguration.TRANSPORT_SOCKET)) {
			
			serviceURL="socket://"+myIP+":"+confPort;
			
		} else  {
			
			String confServicePath=MiddlewareConfiguration.getHttpServicePath();
			serviceURL="http://"+myIP+":"+confPort+confServicePath;
		}

		return serviceURL;
	}
	
	/**
	 * Register a Web service at the SOAP server.
	 * 
	 * @param ws Web service.
	 */
	public synchronized void register(AWebService ws) {
		
		server.registerService(ws.getServiceNamespace(), 
				new WSMethodCallHandler(ws));
		
	}
	
	/**
	 * Deregister a Web service from the SOAP server.
	 * 
	 * @param ws Web service.
	 */
	public synchronized void deregister(AWebService ws) {
		
		server.deregisterService(ws.getServiceNamespace());
		
	}
	
	/**
	 * Returns the SOAP server.
	 * 
	 * @return SOAP server instance.
	 */
	public SOAPServer getServer() {
		return server;
	}
	
	/**
	 * Starts and returnss the local service container.
	 * 
	 * @return Local service container instance.
	 * @throws Exception If the SOAPServer cannot be instantiated.
	 */
	public synchronized static ServiceContainer startLocalServiceContainer() 
	throws Exception {
		
		if (localContainer==null) {
			localContainer=new ServiceContainer();
			localContainer.server.start();
		}
		return localContainer;
	}
	
	/**
	 * Closes the local service container.
	 * 
	 * @throws IOException If the SOAPServer cannot be closed properly.
	 */
	public synchronized static void closeLocalServiceContainer() 
	throws IOException {

		if (localContainer!=null) {
			localContainer.server.stop();
			localContainer=null;
		}
	}
	
	/**
	 * Returns the local service container.
	 * 
	 * @return The local service container or <code>null</code> if none 
	 * is currently running.
	 */
	public synchronized static ServiceContainer getLocalServiceContainer() {
		
		return localContainer;
	}	
}
