/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

import org.ksoap2.serialization.SoapObject;

import at.ac.tuwien.vitalab.middleware.AWebService;

import edu.wustl.mobilab.sliver.soap.SOAPInvocationHandler;
import edu.wustl.mobilab.sliver.soap.SOAPServer;
import edu.wustl.mobilab.sliver.soap.SerializableException;
import edu.wustl.mobilab.sliver.soap.UnserializableException;
import edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext;

/**
 * MethodCaller, able to handle the operation name mapping of
 * AWebService.
 * 
 * @author Greg Hackmann
 * @author Lukasz Juszczyk
 *
 */

public class WSMethodCallHandler implements SOAPInvocationHandler
{
	/**
	 * A mapping from primitive classes to their respective wrapper classes
	 * (e.g., <code>byte.class</code> =>
	 * <code>{@link java.lang.Byte}.class</code>)
	 */
	protected static final Map primitives = new HashMap();
	static
	{
		primitives.put(byte.class, Byte.class);
		primitives.put(short.class, Short.class);
		primitives.put(int.class, Integer.class);
		primitives.put(long.class, Long.class);
		primitives.put(float.class, Float.class);
		primitives.put(double.class, Double.class);
		primitives.put(char.class, Character.class);
		primitives.put(boolean.class, Boolean.class);
		// Add all the primitives to the mapping
	}
	
	/**
	 * The web service that will ultimately handle the incoming requests.
	 */
	protected final AWebService serviceInstance;
	
	/**
	 * Creates a new MethodCallHandler.
	 * 
	 * @param serviceInstance a reference to the web service that will 
	 * handle incoming requests
	 */
	public WSMethodCallHandler(AWebService serviceInstance)
	{
		this.serviceInstance = serviceInstance;
	}
	
	public Object handleInvocation(SoapObject request) throws
		IllegalArgumentException, UnserializableException, SerializableException
	{

		// first see whether we can find a method which handles soap objects
		// directly
		Method method =findMethod(request);	
		Object [] params = null;
		
		if(method == null)
		{
			throw new IllegalArgumentException("Unknown method " +
				request.getName() + " in namespace " + 
				request.getNamespace()+" or invalid parameters.");
		}
		// Try to find the appropriate method
		
		Class[] methodParams=method.getParameterTypes();
		params = new Object[methodParams.length];
		int p=0;
		for(int i = 0; i < params.length; i++)
			if (InvocationContext.class.equals(methodParams[i])) {
				params[i] = SOAPServer.getInvocationContext();
			} else {
				params[i] = request.getProperty(p++);				
			}
		// Copy the request's parameters into an array
			

		
		try
		{
			if (method.getReturnType().equals(void.class)) {
				method.invoke(serviceInstance, params);
				return new SoapResponse(request.getName(),
						serviceInstance.getServiceNamespace());
			} else {
				return new SoapResponse(
						request.getName(),
						serviceInstance.getServiceNamespace(),
						method.invoke(serviceInstance, params));
				// Invoke the method				
			}
		}
		// If the method threw an Exception
		catch(InvocationTargetException e)
		{
			Throwable cause = e.getTargetException();
			// Get the actual Exception that was thrown
			
			if(cause instanceof SerializableException)
				throw (SerializableException)cause;
		
			throw new UnserializableException(e.getMessage(), cause);
			// Re-throw it in the appropriate form
		}
		// If we don't have permission to access the method
		catch(IllegalAccessException e)
		{
			throw new UnserializableException(e.getMessage(), e);
			// Re-throw it in the appropriate form
		}
	}
	
	/**
	 * Finds the method corresponding to an incoming SOAP request.
	 * 
	 * @param request the incoming request
	 * @return the corresponding method, or <code>null</code> if none exists
	 */
	protected Method findMethod(SoapObject request)
	{
		String name = request.getName();
		int numParams = request.getPropertyCount();

		Hashtable nameMapping=serviceInstance.getOperationNameMapping();
		if (nameMapping!=null && nameMapping.containsKey(name)) {
			name=nameMapping.get(name).toString();
		}
		
		Collection nameList=serviceInstance.getOperationNames();
		
		Method [] methods = serviceInstance.getClass().getMethods();

		// For each method
nextMethod:
		for(int i = 0; i < methods.length; i++)
		{
			Method m = methods[i];
		
			if (!m.getName().equals(name) ||
					(!nameMapping.containsValue(m.getName()) &&
					!nameList.contains(m.getName()))) {
				continue;
			}
		
			Class [] paramTypes = m.getParameterTypes();
			if(paramTypes.length < numParams)
				continue;
			
			// For each parameter
			int j = 0;
			int p = 0;
			for(; j < paramTypes.length; j++)
			{
				
				Class paramType = paramTypes[j];
				
				if (InvocationContext.class.equals(paramType)) {
					continue;
				} 
				
				if (p>=numParams) {
					continue nextMethod;
				}
				
				Object param = request.getProperty(p++);

				if(paramType.isPrimitive())
				{
					Class nonPrimitiveType = (Class)primitives.get(paramType);
					if(!nonPrimitiveType.isInstance(param))
						continue nextMethod;
				}
				else if(!paramType.isInstance(param))
					continue nextMethod;
				// Ensure that the method will accept the incoming
				// parameter, converting any primitive types to the appropriate
				// wrapper type first
			}
		
			return m;
		}
		
		return null;
	}
}