/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap.services;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

import edu.wustl.mobilab.sliver.log.Logger;
import edu.wustl.mobilab.sliver.log.LoggerFactory;


import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.MiddlewareConfiguration;
import at.ac.tuwien.vitalab.middleware.discovery.ServiceRegistry;
import at.ac.tuwien.vitalab.middleware.soap.ServiceContainer;
import at.ac.tuwien.vitalab.middleware.soap.WSClassLoader;

/**
 * Deployment Service. Enables hot-deployment of Web services.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class DeploymentService extends AWebService {

	public static final String NAMESPACE=
		MiddlewareConfiguration.getSoapDeploymentNS();
	
	public static final String NAME="DeploymentService";
	
	private static Logger logger=
		LoggerFactory.getLogger(DeploymentService.class);
	
	// key = service class name
	// value = service instance
	private Hashtable services=new Hashtable();
	
	private String repositoryDirectoryPath="repository";
	
	
	/**
	 * Constructor.
	 *
	 */
	public DeploymentService() {
		super(NAME,NAMESPACE);
		
		// we don't publish this service to the outside world
		this.published=false;
	}
	
	/**
	 * Redeploys all persistent services.
	 * 
	 * @throws ClassNotFoundException If the class cannot be found.
	 * @throws IllegalAccessException If the constructor of the Web service
	 * cannot be accessed.
	 * @throws InstantiationException If the Web service cannot be instantiated.
	 * @throws IOException If IO errors occur.
	 */
	public synchronized void redeployPersistentServices() 
	throws ClassNotFoundException, 
	InstantiationException, IllegalAccessException, IOException {
		
		File repoDir=new File(repositoryDirectoryPath);

		if (!repoDir.exists()) {
			return;
		}

		if (!repoDir.isDirectory() || !repoDir.canRead()) {
			logger.error("Cannot access '"+repositoryDirectoryPath+"': " +
			" not a readable directory.");
			return;
		}

		// each subdirectory is one web service
		File[] subDirs=repoDir.listFiles();

		for (int i=0;i<subDirs.length;i++) {
			File serviceDir=subDirs[i];

			if (serviceDir.getName().startsWith(".")) {
				continue;
			}

			if (!serviceDir.isDirectory() || !serviceDir.canRead()) {
				logger.error("Cannot access '"+serviceDir.getAbsolutePath()+
				"':  not a readable directory.");
			}

			Vector jarPaths=new Vector();

			File[] jarFiles=serviceDir.listFiles();

			for (int j=0;j<jarFiles.length;j++) {
				File jar=jarFiles[j];

				if (!jar.isFile() || !jar.canRead()) {
					logger.error("Cannot access '"+
							jar.getAbsolutePath()+"': "+
					" not a readable file.");
				}

				jarPaths.add(jar.getAbsolutePath());
			}

			deployTransient(serviceDir.getName(),jarPaths);
		}
	}
	
	/**
	 * Instantiates a Web service from the class name and optional jar files.
	 * If <code>saveToRepository</code> is set to <code>true</code> the
	 * service will be saved to the repository.
	 * 
	 * @param className Class name of the Web service.
	 * @param jarPaths Optional path to a Jar file which contains the 
	 * Web service.
	 * @param saveToRepository If the Web service should be saved to the
	 * repository.
	 * 
	 * @return Web service instance. 
	 * 
	 * @throws ClassNotFoundException If the class cannot be found.
	 * @throws IllegalAccessException If the constructor of the Web service
	 * cannot be accessed.
	 * @throws InstantiationException If the Web service cannot be instantiated. 
	 * @throws IOException If IO errors occur while saving the service
	 * into the repository.
	 */
	private AWebService instantiateWebService(String className, 
			Vector jarPaths, boolean saveToRepository) 
	throws InstantiationException, IllegalAccessException, 
	IOException, ClassNotFoundException {
		
		// access repository directory
		File repoDir=new File(repositoryDirectoryPath);
		
		File serviceDir=null;
		
		if (saveToRepository) {
			
			if (!repoDir.exists() && !repoDir.mkdir()) {
				throw new IOException(
						"Cannot create '"+repositoryDirectoryPath+"'.");
			}

			if (!repoDir.isDirectory() || !repoDir.canWrite()) {
				throw new IOException("Cannot access '"+
						repositoryDirectoryPath+"': "+
				"not a writeable directory.");
			}

			serviceDir=new File(repoDir,className);

			if (serviceDir.exists()) {
				removeFilesRecursively(serviceDir);
			}

			if (!serviceDir.mkdir()) {
				throw new IOException("Cannot create '"+
						serviceDir.getAbsolutePath()+"'.");
			}
		}
		
		// classloader which can read in new Jar files
		WSClassLoader loader=new WSClassLoader(
				Thread.currentThread().getContextClassLoader());
		
		if (jarPaths!=null && jarPaths.size()>0) {
			for (int i=0;i<jarPaths.size();i++) {
				String path=jarPaths.get(i).toString();
				
				if (path!=null && path.length()>0) {
					File jarFile=new File(path);
					if (jarFile.exists() && jarFile.canRead()) {
						loader.addURL(jarFile.toURL());
						
						// shall the jar be also saved to the repo?
						if (saveToRepository) {
							File repoJarFile=
								new File(serviceDir,jarFile.getName());
							int prefix=0;
							while (repoJarFile.exists()) {
								repoJarFile=new File(serviceDir,
										(prefix++)+
										"_"+jarFile.getName());
							}
							
							copyFile(jarFile,repoJarFile);
						}
					}								
				}
			}
		}
		
		// load the class
		Class serviceClass=loader.loadClass(className);
		AWebService service= (AWebService)(serviceClass.newInstance());
		service.setClassLoader(loader);
		
		return service;
	}
	
	/**
	 * Deploy a Web service  and keep it in the repository.
	 * 
	 * @param className Class name of the Web service.
	 * @param jarPaths Optional path to a Jar file which contains the 
	 * Web service.
	 * @return Dummy boolean since HttpTransport of kSOAP doesn't like
	 * oneWay invocations.
	 * 
	 * @throws ClassNotFoundException If the class cannot be found.
	 * @throws IllegalAccessException If the constructor of the Web service
	 * cannot be accessed.
	 * @throws InstantiationException If the Web service cannot be instantiated. 
	 * @throws IOException If IO errors occur while saving the service
	 * into the repository.
	 */
	public boolean deploy(String className, Vector jarPaths) 
	throws ClassNotFoundException, InstantiationException, 
	IllegalAccessException, IOException {
		boolean ok=false;
		try {
			deploy_(instantiateWebService(className,jarPaths,true));
			ok=true;
		} finally {
			if (!ok) {
				
				File repoDir=new File(repositoryDirectoryPath);
				File serviceDir=new File(repoDir,className);
				try {
					removeFilesRecursively(serviceDir);					
				} catch (IOException ioe) {
					// ignore
				}
			}
		}
		return true;
	}

	
	/**
	 * Deploy a Web service which will undeployed if service container
	 * is shut down.
	 * 
	 * @param className Class name of the Web service.
	 * @param jarPaths Optional path to a Jar file which contains the 
	 * Web service.
	 * @return Dummy boolean since HttpTransport of kSOAP doesn't like
	 * oneWay invocations.
	 * 
	 * @throws ClassNotFoundException If the class cannot be found.
	 * @throws IllegalAccessException If the constructor of the Web service
	 * cannot be accessed.
	 * @throws InstantiationException If the Web service cannot be instantiated. 
	 * @throws IOException If IO errors occur.
	 */
	public boolean deployTransient(String className, Vector jarPaths) 
	throws ClassNotFoundException, InstantiationException, 
	IllegalAccessException, IOException {
		
		deploy_(instantiateWebService(className,jarPaths,false));
		return true;
	}

	/**
	 * Deploy a Web service.
	 * 
	 * @param ws Web service instance.
	 * @return Dummy boolean since HttpTransport of kSOAP doesn't like
	 * oneWay invocations.
	 */
	public boolean deployTransientWS(AWebService ws) {
		deploy_(ws);
		return true;
	}
	
	/**
	 * Deploy a Web service.
	 * 
	 * @param ws Web service.
	 */
	private synchronized void deploy_(AWebService ws) {
				
		ServiceContainer container=ServiceContainer.getLocalServiceContainer();
		ServiceRegistry publisher=ServiceRegistry.getLocalServiceRegistry();
		
		if (container==null) {
			throw new IllegalStateException(
					"Service container not initialized.");
		}
		
		if (publisher==null) {
			throw new IllegalStateException(
					"Service publisher not initialized.");
		}

		String className=ws.getClass().getName();

		if (services.containsKey(className)) {
			throw new IllegalStateException("Service already deployed.");
		}

		Iterator serviceIter=services.values().iterator();
		
		while (serviceIter.hasNext()) {
			AWebService s=(AWebService)serviceIter.next();
			if (s.getServiceNamespace().equals(ws.getServiceNamespace())) {
				throw new IllegalStateException(
						"Service with the same namespaces already deployed.");
			}
		}
		
		services.put(className, ws);

		// register it at the soap server
		container.register(ws);

		// register it at the advertiser
		publisher.publish(ws);				

		try {
			ws.onDeploy();
		} catch (Exception e) {
			logger.warning("Service "+className+" @ "+
					ws.getServiceNamespace()+" threw exception during " +
					"deployment hook: "+e.getClass().getName()+
					": "+e.getMessage());
		}

		logger.info("Service "+className+" @ "+
				ws.getServiceNamespace()+" has been registered.");	
	}
	
	/**
	 * Undeloy a Web service.
	 * 
	 * @param ws Web service instance.
	 * 
	 * @return Dummy boolean since HttpTransport of kSOAP doesn't like
	 * oneWay invocations.
	 */
	public boolean undeployWS(AWebService ws) {
		return undeploy(ws.getClass().getName());
	}
	
	/**
	 * Undeloy a Web service.
	 * 
	 * @param className Class name of the Web service.
	 * 
	 * @return Dummy boolean since HttpTransport of kSOAP doesn't like
	 * oneWay invocations.
	 */
	public synchronized boolean undeploy(String className) {

		ServiceContainer container=ServiceContainer.getLocalServiceContainer();
		ServiceRegistry publisher=ServiceRegistry.getLocalServiceRegistry();
		
		if (container==null) {
			throw new IllegalStateException(
					"Service container not initialized.");
		}
		
		if (publisher==null) {
			throw new IllegalStateException(
					"Service publisher not initialized.");
		}
		
		AWebService ws=(AWebService)services.remove(className);

		if (ws==null) {
			throw new IllegalStateException("Service not deployed.");
		}

		String namespace=ws.getServiceNamespace();

		
		try {
			ws.onUndeploy();				
		} catch (Exception e) {
			logger.warning("Service "+className+" @ "+
					ws.getServiceNamespace()+" threw exception during " +
					"undeployment hook: "+e.getClass().getName()+
					": "+e.getMessage());
		}


		// unregister it from the avertiser
		publisher.unpublish(ws);

		// unregister it from the soap server
		container.deregister(ws);		

		// access repository directory
		File repoDir=new File(repositoryDirectoryPath);


		if (repoDir.exists() && repoDir.isDirectory() && repoDir.canWrite()) {

			File serviceDir=new File(repoDir,className);

			if (serviceDir.exists() && serviceDir.isDirectory() && 
					serviceDir.canWrite()) {

				try {
					removeFilesRecursively(serviceDir);
				} catch (IOException e) {
					logger.error("Cannot delete files of service "+
							className+" @ "+namespace+".");
				}
			}
		}

		logger.info("Service "+className+" @ "+namespace+
		" has been unregistered.");		

		return true;
	}	
	
	/**
	 * Undeploys all services.
	 *
	 */
	public void undeployAllServices() {
		Enumeration servicenameIter=services.keys();
		
		while (servicenameIter.hasMoreElements()) {
			undeploy(servicenameIter.nextElement().toString());
		}
	}
	
	/**
	 * Copies a file.
	 * 
	 * @param in Source.
	 * @param out Destination.
	 * @throws IOException If IO errors occur.
	 */
	private void copyFile(File in, File out) throws IOException {
		
		if (!in.exists()) {
			return;
		}
		
		FileInputStream fis  = new FileInputStream(in);
		FileOutputStream fos = new FileOutputStream(out);
		
		byte[] buf = new byte[64*1024];
		int i = 0;
		
		while((i=fis.read(buf))!=-1) {
			fos.write(buf, 0, i);
		}
		
		fis.close();
		fos.close();
	}
	
	/**
	 * Deletes files recursively.
	 * 
	 * @param f File or directory.
	 * @throws IOException If IO errors occur.
	 */
	protected void removeFilesRecursively(File f) throws IOException {

		if (!f.exists()) {
			return;
		}
		
		if (f.isDirectory()) {
			
			File[] dfArray=f.listFiles();
			for (int i=0;i<dfArray.length;i++) {
				File df=dfArray[i];
				
				removeFilesRecursively(df);
			}
		}
		
		if (!f.delete()) {
			throw new IOException("Cannot delete '"+f.getAbsolutePath()+"'.");
		}
	}

	/**
	 * Returns the path of repository directory.
	 * 
	 * @return Path string.
	 */
	public String getRepositoryDirectoryName() {
		return repositoryDirectoryPath;
	}

	/**
	 * Sets the path of repository directory.
	 * 
	 * @param dir Path string.
	 */
	public void setRepositoryDirectoryName(String dir) {
		repositoryDirectoryPath = dir;
	}


	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.ASingleWebService#getAttributes()
	 */
	public Hashtable getAttributes() {
		return new Hashtable();
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.ASingleWebService#getConfigurationRessourceName()
	 */
	public String getConfigurationRessourceName() {
		return null;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationNames()
	 */
	public Collection getOperationNames() {
		Vector result=new Vector();
		result.add("deploy");
		result.add("undeploy");
		return result;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationDescription(java.lang.String)
	 */
	public Hashtable getOperationDescription(String operationName) {
		if ("deploy".equals(operationName) || 
				"undeploy".equals(operationName)) {
			return new Hashtable();
		} 
		throw new IllegalArgumentException("Invalid operation name.");
	}
	
}
