/*
    This file is part of Vimoware.

    Vimoware is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Vimoware is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with Vimoware. If not, see <http://www.gnu.org/licenses/>.
 */
package at.ac.tuwien.vitalab.middleware.soap.services;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;
import java.util.Vector;

import org.ksoap2.serialization.KvmSerializable;

import edu.wustl.mobilab.sliver.soap.SOAPServer.InvocationContext;

import at.ac.tuwien.vitalab.middleware.AWebService;
import at.ac.tuwien.vitalab.middleware.Middleware;
import at.ac.tuwien.vitalab.middleware.discovery.AttributeRetriever;
import at.ac.tuwien.vitalab.middleware.util.Base64;

/**
 * Info Service. Delivers information about the web services.
 * 
 * @author Lukasz Juszczyk
 *
 */
public class InfoService extends AWebService {

	public static final String NAME="InfoService";
	
	public static final String NAMESPACE=Middleware.NAMESPACE+"info/";
		
	private Hashtable services=new Hashtable();
	
	private int serial=0;
	private int instanceID=0;
	
	private LinkedList changeLog=new LinkedList();
	
	private static HashSet operationNames=new HashSet();
	
	static {
		operationNames.add("getServiceNamespaces");
		operationNames.add("getServiceOperations");
		operationNames.add("getServiceOperationReturnType");
		operationNames.add("getServiceOperationParameterTypes");
		operationNames.add("getServiceAttributes");
		operationNames.add("getServiceOperationDescription");
		operationNames.add("getServiceInfo");
		operationNames.add("getStatusChanges");	
		operationNames.add("getSerializedType");
	}
	
	/**
	 * Constructor.
	 *
	 */
	public InfoService() {
		super(NAME,NAMESPACE);
		instanceID=Math.abs(new Random().nextInt(100000));
	}
	
	/**
	 * Lists all namespaces at which Web services are registered.
	 * 
	 * @return List of namespaces.
	 */
	public Vector getServiceNamespaces() {
		return new Vector(services.keySet());
	}
	
	/**
	 * Returns the number of registered services.
	 * 
	 * @return Number of services.
	 */
	public int getNumberOfServices() {
		return services.size();
	}
	
	/**
	 * Lists all operations of the Web service registered at this namespace.
	 * 
	 * @param namespace Namespace.
	 * @return List of operation names.
	 */
	public Vector getServiceOperations(String namespace) {
		AWebService ws=(AWebService)services.get(namespace);

		if (ws==null) {
			throw new IllegalArgumentException("Illegal/unknown namespace.");
		}						

		Method[] methods=ws.getClass().getMethods();

		HashSet realMethodNames=new HashSet();

		for (int i=0;i<methods.length;i++) {
			Method m=methods[i];
			realMethodNames.add(m.getName());
		}

		Vector result=new Vector();
		
		result.addAll(ws.getOperationNames());
		result.retainAll(realMethodNames);

		Hashtable operationMap=ws.getOperationNameMapping();

		for (Iterator iter=operationMap.keySet().iterator(); iter.hasNext();) {

			String opName=iter.next().toString();
			if (realMethodNames.contains(operationMap.get(opName))) {
				result.add(opName);
			}
		}
		
		return result;		
	}
	
	/**
	 * Delivers the return type of the operation of the Web service registered
	 * at this namespace.
	 * 
	 * @param namespace Namespace.
	 * @param operation Operation name.
	 * @return Return type.
	 */
	public String getServiceOperationReturnType(String namespace, 
			String operation) {
		
		AWebService ws=(AWebService)services.get(namespace);

		if (ws==null) {
			throw new IllegalArgumentException("Illegal/unknown namespace.");
		}

		Method[] methods=ws.getClass().getMethods();

		Hashtable mapping=ws.getOperationNameMapping();
		if (mapping!=null && mapping.containsKey(operation)) {
			operation=mapping.get(operation).toString();
		}

		for (int i=0;i<methods.length;i++) {
			Method m=methods[i];

			if (m.getName().equals(operation)) {
				return m.getReturnType().getName();

			}
		}

		throw new IllegalArgumentException("Illegal/unknown operation.");
	}
	
	/**
	 * Lists all parameter types of the operation of the Web service registered
	 * at this namespace.
	 * 
	 * @param namespace Namespace.
	 * @param operation Operation name.
	 * @return List of parameter types.
	 */
	public Vector getServiceOperationParameterTypes(
			String namespace, String operation) {
		
		AWebService ws=(AWebService)services.get(namespace);

		if (ws==null) {
			throw new IllegalArgumentException("Illegal/unknown namespace.");
		}

		Method[] methods=ws.getClass().getMethods();

		Hashtable mapping=ws.getOperationNameMapping();
		if (mapping!=null && mapping.containsKey(operation)) {
			operation=mapping.get(operation).toString();
		}

		for (int i=0;i<methods.length;i++) {
			Method m=methods[i];

			if (m.getName().equals(operation)) {

				Vector result=new Vector();

				Class[] params=m.getParameterTypes();

				for (int j=0;j<params.length;j++) {
					if (!params[j].equals(InvocationContext.class)) {
						result.add(params[j].getName());
					}
				}

				return result;

			}
		}

		throw new IllegalArgumentException("Illegal/unknown operation.");
	}
		
//	/**
//	 * Returns a message type in a serialized form.
//	 * 
//	 * @param type Data type name.
//	 * @return Serialized type.
//	 * @throws IOException If the type is invalid or errors occur.
//	 */
//	public String getSerializedType(String type) throws IOException {
//		
//		Class clazz;
//		
//		try {
//			clazz=Class.forName(type);
//			if (!KvmSerializable.class.isAssignableFrom(clazz)) {
//				throw new IOException("Invalid type '"+type+"'" +": " +
//						"no KvmSerializable.");
//			}
//			
//		} catch (ClassNotFoundException e) {
//			throw new IOException("Unknown/invalid type '"+type+"'.");			
//		}
//		
//		InputStream is=Thread.currentThread().
//				getContextClassLoader().
//				getResourceAsStream(clazz.getName().replace('.', '/')+
//						".class");
//				
//		is=new BufferedInputStream(is);
//		
//		ByteArrayOutputStream baos=new ByteArrayOutputStream();
//		int len=-1;
//		int buffSize=1024;
//		byte[] buffer=new byte[buffSize];
//		
//		while ((len =is.read(buffer, 0, buffSize))!=-1) {
//			baos.write(buffer, 0, len);
//		}
//		
//		String ser=Base64.encodeBytes(baos.toByteArray(),
//				Base64.DONT_BREAK_LINES);
//		
//		return ser;
//	}
	
	/**
	 * Returns all attributes of the operation of the Web service registered
	 * at this namespace.
	 * 
	 * @param namespace Namespace.
	 * @param operation Operation name.
	 * @return Hashtable of attributes.
	 */
	public Hashtable getServiceOperationDescription(
			String namespace, String operation) {
		
		AWebService ws=(AWebService)services.get(namespace);

		if (ws==null) {
			throw new IllegalArgumentException("Illegal/unknown namespace.");
		}

		Method[] methods=ws.getClass().getMethods();

		Hashtable mapping=ws.getOperationNameMapping();
		if (mapping!=null && mapping.containsKey(operation)) {
			operation=mapping.get(operation).toString();
		}

		for (int i=0;i<methods.length;i++) {
			Method m=methods[i];

			if (m.getName().equals(operation)) {

				return ws.getOperationDescription(operation);

			}
		}

		throw new IllegalArgumentException("Illegal/unknown operation.");
	}
	
	/**
	 * Returns all attributes of a service.
	 * 
	 * @param namespace Namespace.
	 * @return Hashtable of attributes.
	 */
	public Hashtable getServiceAttributes(String namespace) {
		
		AWebService ws=(AWebService)services.get(namespace);

		if (ws==null) {
			throw new IllegalArgumentException("Illegal/unknown namespace.");
		}
		return AttributeRetriever.getAttributes(ws);
	}
	
	/**
	 * Returns all information about a service.
	 * 
	 * @param namespace Namespace.
	 * @return Hashtable containing the information.
	 */
	public Hashtable getServiceInfo(String namespace) {

		Hashtable result=new Hashtable();
		
		Collection operationNames=getServiceOperations(namespace);
		Iterator operationIter=operationNames.iterator();
		
		result.put("attributes", getServiceAttributes(namespace));
		
		while (operationIter.hasNext()) {
			
			String operationName=operationIter.next().toString();
			
			Hashtable serviceInfo=new Hashtable();
			
			result.put("op."+operationName, serviceInfo);
			
			serviceInfo.put("return",
					getServiceOperationReturnType(namespace,operationName));
			
			serviceInfo.put("params",
					getServiceOperationParameterTypes(namespace,operationName));
			
			try {
				Hashtable description=
					getServiceOperationDescription(namespace, operationName);
				if (description!=null && !description.isEmpty()) {
					serviceInfo.put("description", description);					
				}
			} catch (Exception e) {
				// ignore
			}
		}
		
		return result;
	}
	
	/**
	 * Returns the changes in the status of the host.
	 * 
	 * @param fromSerial Last serial number retrieved by the client.
	 * @param toSerial Current serial number received by the client.
	 * @return Vector with two lists. First with all service namespaces
	 * which have been added since then. Second with all service namespaces
	 * which have been removed since then. 
	 */
	public synchronized Vector getStatusChanges(int fromSerial, int toSerial) {
		
		HashSet minus=new HashSet();
		HashSet plus=new HashSet();

		if (toSerial>serial) {
			toSerial=serial;
		}
		
		if (fromSerial<0) {
			fromSerial=0;
		}
			
		if (fromSerial==0 && toSerial==serial) {
			
			plus.addAll(getServiceNamespaces());
			
		} else {
			Iterator logIter=null;
			try {				
				logIter=new Vector(changeLog).listIterator(fromSerial);				
			} catch (IndexOutOfBoundsException e) {
				throw new IllegalArgumentException(e.getMessage());
			}
			
			while (logIter.hasNext() && (fromSerial++<toSerial)) {
				Object[] tuple=(Object[])logIter.next();
				AWebService ws=(AWebService)tuple[0];
				String namespace=ws.getServiceNamespace();
				String plusOrMinus=tuple[1].toString();
				
				if ("+".equals(plusOrMinus)) {
					if (minus.contains(namespace)) {
						minus.remove(namespace);
					} else {
						plus.add(namespace);
					}
				} else {
					if (plus.contains(namespace)) {
						plus.remove(namespace);
					} else {
						minus.add(namespace);
					}
				}
			}
		}
		
		Vector result=new Vector();
		result.add(new Vector(plus));
		result.add(new Vector(minus));
		
		return result;
	}

	/**
	 * Retures the last status change as a String.
	 * 
	 * @return Last change.
	 */
	public synchronized String getLastStatusChange() {
		
		if (changeLog.isEmpty()) {
			return "";			
		}
		
		Object[] last=(Object[])changeLog.getLast();
		
		return last[1]+((AWebService)last[0]).getServiceNamespace();
	}
	
	/**
	 * Register a Web service.
	 * 
	 * @param ws Web service.
	 */
	public synchronized void registerService(AWebService ws) {
		serial++;
		changeLog.add(new Object[]{ws,"+"});
		services.put(ws.getServiceNamespace(), ws);		
	}
	
	/**
	 * Unregister a Web service.
	 * 
	 * @param ws Web service.
	 * @return <code>true</code> if the service existed, 
	 * otherwise <code>false</code>.
	 */
	public synchronized boolean unregisterService(AWebService ws) {
		serial++;
		changeLog.add(new Object[]{ws,"-"});
		return services.remove(ws.getServiceNamespace())!=null;			
	}

	/**
	 * Returns the serial number.
	 * 
	 * @return Serial number.
	 */
	public int getSerial() {
		return serial;
	}

	/**
	 * Returns the instance ID.
	 * 
	 * @return Instance ID.
	 */
	public int getInstanceID() {
		return instanceID;
	}
	
	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getAttributes()
	 */
	public Hashtable getAttributes() {
		return new Hashtable();
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getConfigurationRessourceName()
	 */
	public String getConfigurationRessourceName() {
		return null;
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationNames()
	 */
	public Collection getOperationNames() {
		return new Vector(operationNames);
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationNameMapping()
	 */
	public Hashtable getOperationNameMapping() {
		return new Hashtable();
	}

	/* (non-Javadoc)
	 * @see at.ac.tuwien.vitalab.middleware.AWebService#getOperationDescription(java.lang.String)
	 */
	public Hashtable getOperationDescription(String operationName) {
		if (operationNames.contains(operationName)) {
			return new Hashtable();
		} 
		throw new IllegalArgumentException("Invalid/unknown operation.");		
	}
	
}
